"use client";
import {
    createContext,
    ReactNode,
    SetStateAction,
    useContext,
    useEffect,
    useRef,
    useState,
} from "react";

import { ID, PageData } from "@snip/database/types";

import { useSocketContext } from "../useSocketContext";

interface PagesContext {
    pages: Map<ID, PageData>;
    pagesRef: React.MutableRefObject<Map<ID, PageData>>;
    pagesArray: PageData[];
    createPage: () => Promise<PageData>;
    updatePage: (page: PageData) => void;
    deletePage: (page_id: ID) => void;
}

const PagesContext = createContext<PagesContext | null>(null);

/** Has to be nested inside a socket context to work
 * properly.
 */
export function PagesContextProvider({
    init_pages,
    children,
}: {
    children: ReactNode;
    init_pages: PageData[];
}) {
    const { socket, isConnected, book_id } = useSocketContext();

    /** The pages of the book
     * updates with creation or deletion from the server
     */
    const [pages, _setPages] = useState<Map<ID, PageData>>(
        init_pages.reduce((map, page) => {
            map.set(page.id, page);
            return map;
        }, new Map()),
    );
    const pagesRef = useRef(pages);

    const setPages = (pages: SetStateAction<Map<ID, PageData>>) => {
        if (pages instanceof Function) {
            pagesRef.current = pages(pagesRef.current);
        }
        _setPages(pagesRef.current);
    };
    // Register event listeners
    useEffect(() => {
        if (!isConnected || !socket) return;
        function onNewPage(page: PageData) {
            page = parsePageData(page);
            setPages((pages) => {
                pages.set(page.id, page);
                return new Map(pages);
            });
        }
        function onDeletedPage(page_id: ID) {
            setPages((pages) => {
                pages.delete(page_id);
                return new Map(pages);
            });
        }
        function onUpdatedPage(page: PageData) {
            page = parsePageData(page);
            setPages((pages) => {
                pages.set(page.id, page);
                return new Map(pages);
            });
        }
        function onUpdatedPagePreview(page_id: ID) {
            setPages((pages) => {
                const page = pages.get(page_id);
                if (!page) return pages;
                page.last_updated = new Date();
                return new Map(pages);
            });
        }

        // Page events
        socket.on("book:newPage", onNewPage);
        socket.on("book:deletedPage", onDeletedPage);
        socket.on("book:updatePage", onUpdatedPage);
        socket.on("book:updatePage:preview", onUpdatedPagePreview);

        return () => {
            socket.off("book:newPage", onNewPage);
            socket.off("book:deletedPage", onDeletedPage);
            socket.off("book:updatePage", onUpdatedPage);
            socket.off("book:updatePage:preview", onUpdatedPagePreview);
        };
    }, [socket, isConnected]);

    async function createPage() {
        if (!book_id || !socket) {
            throw new Error("Cant create page without book set!");
        }
        let newPage: PageData = {
            id: -Math.round(Math.random() * 1000),
            page_number: null,
            book_id: book_id,
            created: new Date(),
            last_updated: new Date(),
        };

        // Send the new page to the server
        newPage = await new Promise((resolve, reject) => {
            socket.emit("book:newPage", newPage, (newPage: PageData) => {
                if (!newPage) {
                    reject("Could not create page");
                }
                resolve(newPage);
            });
        });
        setPages((pages) => {
            pages.set(newPage.id, newPage);
            return new Map(pages);
        });
        return newPage;
    }

    return (
        <PagesContext.Provider
            value={{
                pages,
                pagesArray: Array.from(pages.values()),
                createPage,
                updatePage: (page) => {
                    if (!socket) return;
                    setPages((pages) => {
                        pages.set(page.id, page);
                        return new Map(pages);
                    });
                    socket.emit("book:updatePage", page, (page: PageData) => {
                        page = parsePageData(page);
                        setPages((pages) => {
                            pages.set(page.id, page);
                            return new Map(pages);
                        });
                    });
                },
                deletePage: (page_id) => {
                    if (!socket) return;
                    setPages((pages) => {
                        pages.delete(page_id);
                        return new Map(pages);
                    });
                    socket.emit("book:deletePage", page_id);
                },
                pagesRef,
            }}
        >
            {children}
        </PagesContext.Provider>
    );
}

export function usePagesContext() {
    const context = useContext(PagesContext);
    if (!context) throw new Error("Cant use PagesContext outside provider!");
    return context;
}

// Reducers
export function usePagesRef() {
    const { pagesRef } = usePagesContext();
    return pagesRef;
}

function parsePageData(page: PageData) {
    page.created = page.created ? new Date(page.created) : new Date();
    page.last_updated = page.last_updated
        ? new Date(page.last_updated)
        : new Date();
    return page;
}
