"use client";

import {
    createContext,
    useCallback,
    useContext,
    useEffect,
    useMemo,
    useState,
} from "react";

import { BaseData, BaseView, ID, SnipData } from "@snip/database/types";
import { get_snip_from_data_allow_unknown } from "@snip/snips";
import { BaseSnip } from "@snip/snips/general/base";
import { ImageSnip } from "@snip/snips/general/image";
import { ImageSnipLegacy } from "@snip/snips/general/legacy/image";

import { useSocketContext } from "../useSocketContext";

interface QueuedSnipsContext {
    snips: Map<ID, BaseSnip>;
    snipsArray: BaseSnip[];
    addTempSnip: (snip: BaseSnip) => void;
    removeQueuedSnip: (id: number) => void;
}

const QueuedSnipsContext = createContext<QueuedSnipsContext | null>(null);

export function QueuedSnipsContextProvider({
    init_snips,
    children,
}: {
    children: React.ReactNode;
    init_snips: SnipData<BaseData, BaseView>[];
}) {
    const { socket, isConnected } = useSocketContext();

    /** The pages of the book
     * updates with creation or deletion from the server
     */
    const [snips, setSnips] = useState<Map<ID, BaseSnip>>(
        init_snips.reduce((map, snipData) => {
            const snip = loadSnip(snipData);
            map.set(snip.id, snip);
            return map;
        }, new Map()),
    );

    // Also on placed snip
    function removeQueuedSnip(snip_id: ID) {
        setSnips((snips) => {
            snips.delete(snip_id);
            return new Map(snips);
        });
    }

    //Register event listeners
    useEffect(() => {
        if (!isConnected || !socket) return;

        async function onNewSnip(snipData: SnipData<BaseData, BaseView>) {
            const snip = loadSnip(snipData);
            if (!snip) return;

            setSnips((snips) => {
                snips.set(snip.id, snip);
                return new Map(snips);
            });
        }

        socket.on("book:newQueuedSnip", onNewSnip);
        socket.on("book:deleteQueuedSnip", removeQueuedSnip);

        return () => {
            socket.off("book:newQueuedSnip", onNewSnip);
            socket.off("book:deleteQueuedSnip", removeQueuedSnip);
        };
    }, [socket, isConnected]);

    const snipsArray = useMemo(() => {
        const array = Array.from(snips.values());
        array
            .sort((a, b) => a.created.getDate() - b.created.getDate())
            .reverse();
        return array;
    }, [snips]);

    const addTempSnip = useCallback((snip: BaseSnip) => {
        if (snip.id > 0) {
            throw new Error("Cant add a temporary snip with a potential DB id");
        }
        setSnips((snips) => {
            snips.set(snip.id, snip);
            return new Map(snips);
        });
    }, []);

    return (
        <QueuedSnipsContext.Provider
            value={{
                snips,
                snipsArray,
                addTempSnip,
                removeQueuedSnip,
            }}
        >
            {children}
        </QueuedSnipsContext.Provider>
    );
}

export function useQueuedSnipsContext() {
    const context = useContext(QueuedSnipsContext);
    if (!context) {
        throw new Error(
            "useQueuedSnipsContext must be used within a QueuedSnipsContextProvider",
        );
    }
    return context;
}

function loadSnip(snipData: SnipData<BaseData, BaseView>) {
    let snip = get_snip_from_data_allow_unknown(snipData);
    if (snip instanceof ImageSnipLegacy) {
        snip = ImageSnip.from_legacy(snip);
    }
    return snip;
}
