import EventEmitter from "events";
import { Socket } from "socket.io-client";

import { BaseData, BaseView, SnipData } from "@snip/database/types";
import { render_background_by_id } from "@snip/render/backgrounds";
import { LocalPage } from "@snip/render/page";
import { ConditionalRender } from "@snip/render/types";
import { get_snip_from_data_allow_unknown } from "@snip/snips";
import { ArraySnip } from "@snip/snips/general/array";

import { getSocket } from "components/context/socket";
export class Page extends LocalPage {
    public socket?: Socket;

    /** Event emitter for all changes to the page.
     *
     * The idea is to e.g. subscribe to the change event and
     * rerender the page.
     */
    public events = new EventEmitter();

    async init(): Promise<void> {
        this.socket = await getSocket(this.data.book_id);
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        if (!this.socket.connected) {
            await this.socket.connect();
        }
        await this.joinRoom();
        await this.subscribeToChanges();
    }

    /** Get the preview image of the page */
    public async getPreviewImageBuffer(): Promise<Buffer> {
        return await new Promise<Buffer>((resolve, reject) => {
            if (!this.socket) return reject();
            this.socket.emit(
                "page:getPreviewImage",
                this.data.id,
                (buffer: Buffer) => {
                    resolve(buffer);
                },
            );
        });
    }

    /** Join room for the page */
    private async joinRoom(): Promise<void> {
        return await new Promise<void>((resolve, reject) => {
            if (!this.socket) return reject();
            this.socket.emit("page:join", this.data.id, (success: boolean) => {
                if (success) {
                    resolve();
                } else {
                    reject();
                }
            });
        });
    }

    /** Use socket to load all snips */
    public async loadAllSnips(): Promise<void> {
        return new Promise<void>((resolve, reject) => {
            if (!this.socket) return reject();
            this.socket.emit(
                "page:getSnips",
                this.data.id,
                (snips: SnipData<BaseData, BaseView>[]) => {
                    this.set_snips(snips.map(get_snip_from_data_allow_unknown));
                    resolve();
                },
            );
        });
    }

    /** Images need
     * to be loaded before they can be rendered.
     */
    public async allSnipsPrepared(): Promise<boolean> {
        const snips_with_ready = this.snips.filter((snip) => {
            return Object.hasOwn(snip, "ready");
        });
        return Promise.all(snips_with_ready.map((snip) => snip.ready)).then(
            () => true,
        );
    }

    /** Subscribe to changes */
    private subscribeToChanges(): void {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        this.socket.on(
            "snip:insert",
            (newSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (newSnip.page_id === this.data.id) {
                    this.add_snip(get_snip_from_data_allow_unknown(newSnip));
                    this.events.emit("insert", newSnip);
                }
            },
        );

        this.socket.on(
            "snip:update",
            (updatedSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (updatedSnip.page_id === this.data.id) {
                    this.update_snip(
                        get_snip_from_data_allow_unknown(updatedSnip),
                    );
                    this.events.emit("update", updatedSnip);
                }
            },
        );

        this.socket.on("snip:remove", (snipId: number) => {
            // Check if snip is on this page
            const success = this.remove_snip(snipId);
            if (success) {
                this.events.emit("remove", snipId);
            }
        });

        this.socket.on("snip:placed", (pageId: number, snipId: number) => {
            // Check if snip is on this page
            if (pageId === this.data.id) {
                // Check if array snip and if so, update the snip
                const snip = this.snips.find((s) => s.id === snipId);
                if (snip && (snip as ArraySnip).snips) {
                    // find all snips in array and remove them
                    const array_snip = snip as ArraySnip;
                    const snips_to_remove = array_snip.snips.map((s) => s.id);
                    // find in snips
                    this.snips = this.snips.filter(
                        (s) => !snips_to_remove.includes(s.id),
                    );
                }
                this.events.emit("placed", snipId);
            }
        });
    }

    /** Render a page
     *
     * @param ctx Context to render onto
     */
    public render(
        ctx: CanvasRenderingContext2D | OffscreenCanvasRenderingContext2D,
        conditionalRender?: ConditionalRender,
    ): void {
        // Render background
        render_background_by_id(this.data.background_type_id, ctx, this.size);

        const sorted_snips = this.snips.sort(
            (a, b) => a.last_updated.getTime() - b.last_updated.getTime(),
        );

        for (const snip of sorted_snips) {
            // Eraser mode decreases opacity of
            // snips older than x seconds
            if (conditionalRender && conditionalRender.condition(snip)) {
                conditionalRender.render(ctx, snip);
            } else {
                snip.render(ctx);
            }
        }
    }

    get id() {
        return this.data.id;
    }
}
