"use client";
import { type UserRouteResponseReturn } from "app/api/account/user/route";
import useSWR, { KeyedMutator, SWRConfiguration } from "swr";

import {
    GroupData,
    GroupDataWithMembers,
    UserConfigData,
    UserData,
} from "@snip/database/types";
import { debounceWithReturn } from "@snip/utils";

import { fetcher } from "../fetcher";
import { use, useEffect } from "react";

interface UserHook {
    user: UserData | undefined;
    config: UserConfigData | undefined;

    isLoading: boolean;
    isError: boolean;
    isLoggedIn: boolean;
    isAdmin: boolean;

    updateUser: (
        userData?: UserData,
        userConfig?: Partial<UserConfigData> & { id: number },
        userGroups?: GroupData[],
    ) => Promise<void>;
    mutate: KeyedMutator<UserRouteResponseReturn>;
}

export default function useUser(config = true) {
    const swrConfig: SWRConfiguration = {
        revalidateIfStale: true,
        revalidateOnFocus: true,
        revalidateOnReconnect: true,
        refreshInterval: 0,
        fallbackData: {
            user: userDataFallback,
            userConfig: config && userConfigFallback,
        },
    };

    const fetcher = async (url: string) => {
        const data = (await fetcher(url)) as UserRouteResponseReturn;

        //Parse dates
        data.user.created = new Date(data.user.created!);
        data.user.last_updated = new Date(data.user.last_updated!);
        data.user.last_login = new Date(data.user.last_login!);

        if (data.userConfig) {
            data.userConfig.last_updated = new Date(
                data.userConfig.last_updated!,
            );
        }
        if (data.groups) {
            data.groups.forEach((g: GroupData | GroupDataWithMembers) => {
                g.created = new Date(g.created!);
                g.last_updated = new Date(g.last_updated!);

                // Iterate once to add the self flag
                if ("members" in g) {
                    g.members.forEach((m) => {
                        m.joined_at = new Date(m.joined_at!);
                        if (m.user_id === data.user.id) {
                            m.self = true;
                        }
                    });
                }
            });
        }

        return data as UserRouteResponseReturn;
    };

    const { data, mutate, error, isLoading } = useSWR<UserRouteResponseReturn>(
        `/api/account/user?config=${config}`,
        fetcher,
        swrConfig,
    );

    useEffect(() => {
        console.log(data)
    }, [data]);


    const updateServerUserConfigDebounced = debounceWithReturn(
        updateServerUserConfigData,
        500,
    );

    const updateUser = async (
        userData?: UserData,
        userConfig?: Partial<UserConfigData> & { id: number },
        userGroups?: GroupData[] | GroupDataWithMembers[],
    ) => {
        if (userData || userGroups) {
            console.warn(
                "Setting userData or userGroups is not supported yet! And might not be supported in the future",
            );
            return;
        }
        if (userConfig) {
            const optimistic = {
                ...data,
            };

            optimistic.userConfig = {
                ...optimistic.userConfig,
                ...userConfig,
            } as UserConfigData;

            // optimistic update
            await mutate(
                async () => {
                    const newConfig =
                        await updateServerUserConfigDebounced(userConfig);

                    optimistic.userConfig = newConfig;

                    return optimistic as UserRouteResponseReturn;
                },
                {
                    optimisticData: optimistic as UserRouteResponseReturn,
                    rollbackOnError: true,
                    populateCache: true,
                    revalidate: false,
                },
            );
        }
    };

    const ret: UserHook = {
        user: data?.user,
        config: data?.userConfig,

        isLoading,
        isError: error,

        isLoggedIn:
            !isLoading &&
            !error &&
            data?.user.id !== undefined &&
            data?.user.id > 0,
        isAdmin: data?.groups?.some((g) => g.name === "admin") || false,

        updateUser,
        mutate,
    };
    return ret;
}

async function updateServerUserConfigData(
    configUpd: Partial<UserConfigData> & { id: number },
): Promise<UserConfigData> {
    return await fetch("/api/account/config", {
        method: "POST",
        headers: {
            "Content-Type": "application/json",
        },
        body: JSON.stringify(configUpd),
    })
        .then((res) => res.json())
        .then((res) => {
            if (res.error) {
                console.error(res);
                throw new Error(res.error);
            }
            return res;
        });
}

/** THIS API SHOULD NOT BE USED FOR CRITICAL APPLICATION PARTS
 *  AS A RESPONSE CAN BE MANIPULATED BY THE USER EASILY!
 *
 * Only use this API to show or hide certain parts of the UI
 *
 */
export function useIsAdmin() {
    const {
        data: isAdmin,
        mutate: mutateIsAdmin,
        error,
    } = useSWR<boolean>("/api/authentication/isAdmin", fetcher);

    return {
        isAdmin: isAdmin,
        isLoading: !error && !isAdmin,
        isError: error,
        mutateIsAdmin,
    };
}

/* -------------------------------------------------------------------------- */
/*                                fallback data                               */
/* -------------------------------------------------------------------------- */
const userDataFallback: UserData = {
    created: new Date(),
    email: "",
    emailValidated: false,
    id: -1,
    last_updated: new Date(),
};
const userConfigFallback: UserConfigData = {
    id: -1,
    last_updated: new Date(),
    pen_color: "#000000",
    pen_colors: ["#000000", "#ff0000", "#00ff00", "#0000ff"],
    pen_size: 4,
    pen_smoothing: 0.0,
    text_font_id: 1,
    text_font: "Arial",
    text_fontColor: "#000000",
    text_fontSize: 12,
    text_lineHeight: 1.25,
    text_lineWrap: 400,
    user_id: -1,
    zoom1: 1,
    zoom2: 1,
};
