import "./overrides";

import { createCanvas, DOMMatrix, registerFont } from "canvas";
import glob from "fast-glob";
import { createReadStream, createWriteStream, existsSync, mkdirSync } from "fs";
import path from "path";
import sharp from "sharp";
import { Readable } from "stream";
import { pipeline } from "stream/promises";

import { config } from "@snip/config/server";
import service from "@snip/database";
import { FONT_URLS } from "@snip/database/fonts/internal";
import { LocalPage } from "@snip/render/page";
import { get_snip_from_data } from "@snip/snips";
import {
    UnknownNamespaceError,
    UnknownSnipTypeError,
} from "@snip/snips/errors";
import { DummySnip } from "@snip/snips/general/dummy";

import { log } from "../logging";
import { type RenderParams } from "./renderParams";

// We need to load fonts for the page
// this can be problematic for tsc
// so we have to use a config path
if (process.env.NODE_ENV !== "test") {
    for (const [name, family] of FONT_URLS) {
        registerFont(path.join("@assets/fonts/", name), {
            family: family,
        });
    }
}

/** Partially public render function
 * allows to render a page by its id
 */
export default async function renderPage(
    page_id: number,
    params: RenderParams,
): Promise<Readable | true> {
    /** Check if file exists already everything with a higher format
     * than the requested one will be rendered
     *
     * /renders/{page_id}/{width}.{format}
     */
    const dir = `${config.RENDER.DIR}/${page_id}`;

    let transform_webp = false;
    if (params.format === "webp") {
        transform_webp = true;
        params.format = "jpeg";
    }

    let stream: Readable | undefined;
    let fs_stream: Promise<true> = Promise.resolve(true);
    if (!params.force_render) {
        stream = await pageFromFile(page_id, params.width, params.format);
    }

    if (stream === undefined) {
        stream = await pageFromDB(page_id, params.width, params.format);
        // Write the stream to the file system
        if (!existsSync(dir)) {
            mkdirSync(dir, {
                recursive: true,
            });
        }
        fs_stream = pipeline(
            stream,
            createWriteStream(`${dir}/${params.width}.${params.format}`),
            {
                end: false,
            },
        ).then(() => true);
    }

    if (!params.return_stream) {
        return await fs_stream;
    }

    // Transform to webp if requested
    if (transform_webp) {
        params.format = "webp";
        const webp_transform = sharp().webp();
        stream = stream.pipe(webp_transform);
    }

    return stream;
}

/** Helper function to find a rendered file
 * if it exists
 *
 * @param page_id the id of the page
 * @param width the width of the file
 * @param format the format of the file
 * @returns a readable stream if the file exists undefined otherwise
 */
async function pageFromFile(
    page_id: number,
    width: number,
    format: string,
): Promise<Readable | undefined> {
    const dir = `${config.RENDER.DIR}/${page_id}`;
    const files = await glob.glob(`${dir}/*.${format}`);
    for (const file of files) {
        const [fileWidth] = file.split("/").pop()!.split(".");
        if (width === parseInt(fileWidth!)) {
            return createReadStream(file);
        }
    }
}

/**
 * Renders a page from the database and returns the stream of the rendered page.
 * @param page_id - The ID of the page to render.
 * @param width - The width of the rendered page.
 * @param format - The format of the rendered page, which can be "pdf", "jpeg", or "png".
 * @returns A Promise that resolves to a Readable stream of the rendered page, or undefined if the page data is not found.
 */
async function pageFromDB(
    page_id: number,
    width: number,
    format: "pdf" | "jpeg" | "png",
): Promise<Readable> {
    /** Get page data
     * including all snips
     */
    const pageData = await service.page.getById(page_id);
    const page = LocalPage.from_data(pageData);
    const snipsData = await service.snip.getByPageId(page_id);
    const snips = snipsData.map((snipData) => {
        try {
            return get_snip_from_data(snipData);
        } catch (e) {
            if (e instanceof UnknownSnipTypeError) {
                log("Unknown snip type", e.message);
            } else if (e instanceof UnknownNamespaceError) {
                log("Unknown namespace", e.message);
            } else if (e instanceof Error) {
                log("Unknown error", e.message);
            } else {
                log("Unknown error", e);
            }
        }
        return new DummySnip(snipData);
    });

    // Prepare snip images
    const promises: Promise<boolean>[] = [];
    for (const snip of snips) {
        if (Object.hasOwn(snip, "ready")) {
            promises.push(snip.ready!);
        }
    }
    await Promise.all(promises);
    page.set_snips(snips);

    /** Render the page
     * write the stream to the file system
     * and return the stream
     */
    const stream = render(page, width, format);
    return stream;
}

/** Helper function to render a local page
 *  expects a fully initialized local page!
 */
export function render(
    localPage: LocalPage,
    width: number,
    format: "pdf" | "jpeg" | "png",
): Readable {
    const canvas = createCanvas(
        width,
        (width * 2000) / 1400,
        format != "pdf" ? undefined : "pdf",
    );

    const ctx = canvas.getContext("2d");
    const transform = ctx.getTransform().invertSelf();
    const t_width =
        transform.a * canvas.width + transform.b * canvas.height + transform.e;
    const t_height =
        transform.c * canvas.width + transform.d * canvas.height + transform.f;
    ctx.fillStyle = "white";
    ctx.fillRect(0, 0, t_width, t_height);

    const scale = canvas.height / 2000;
    const t = new DOMMatrix([scale, 0, 0, scale, 0, 0]);
    ctx.setTransform(t);
    localPage.render(ctx);

    switch (format) {
        case "pdf":
            return canvas.createPDFStream();
        case "jpeg":
            return canvas.createPNGStream();
        case "png":
            return canvas.createJPEGStream({ quality: 0.6 });
    }
}
