import bcrypt from "bcryptjs";

import { registerService } from "@snip/utils";

import { check_mariadb_warning, NotFoundError } from "../errors";
import { pool_permissions as pool } from "../sql.connection";
import { UserStrategy } from "../strategies";
import { ID, PartialBy, UserData } from "../types";
import { groupService } from "./group";

export class UserService implements UserStrategy {
    async getAll(): Promise<UserData[]> {
        const sql =
            "SELECT id, email, emailValidated, created, last_updated FROM users";
        const users: UserData[] = await pool
            .query(sql)
            .then(check_mariadb_warning)
            .then((res) => {
                return res;
            });

        return users;
    }

    /**
     * Inserts a new user into the database. Needs a hashed password set in the user object.
     *
     * @param user - The user data to be inserted.
     * @returns A promise that resolves to the inserted user data, or undefined if the insertion fails.
     */
    async insert(
        user: PartialBy<UserData, "id"> & { hashedPassword: string },
    ): Promise<UserData> {
        //Stringify group ids
        const sql =
            "INSERT INTO users (`email`,  `emailValidated`, `hashedPassword`) VALUES (?, ?, ?)";

        const user_id: ID = await pool
            .query(sql, [
                user.email,
                user.emailValidated || 0,
                user.hashedPassword,
            ])
            .then(check_mariadb_warning)
            .then((res) => {
                return res.insertId;
            });

        user.id = user_id;
        user.created = new Date();
        user.last_updated = new Date();

        // Delete hashedPassword
        delete user.hashedPassword;

        return user as UserData;
    }

    /**
     * Retrieves a user by their ID.
     * @param id - The ID of the user.
     * @returns A promise that resolves to the user data if found, otherwise undefined.
     */
    async getById(id: ID): Promise<UserData> {
        const sql =
            "SELECT id, email, emailValidated, created, last_updated, last_login, last_login_ip FROM users WHERE id = ?";

        const user = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new Error(
                        "Number of users does not match! length: " +
                            res.length +
                            " for id: " +
                            id,
                    );
                }
                return res[0];
            });

        return user;
    }

    /**
     * Retrieves user data for the given array of IDs.
     * @param ids - An array of user IDs.
     * @returns A promise that resolves to an array of user data.
     */
    async getByIds(ids: ID[]): Promise<UserData[]> {
        const sql =
            "SELECT id, email, emailValidated, created, last_updated, last_login, last_login_ip FROM users WHERE id IN (?)";

        const users = await pool
            .query(sql, [ids])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != ids.length) {
                    throw new Error(
                        "Number of Users does not match! length:" +
                            res.length +
                            " for ids: " +
                            ids,
                    );
                }

                return res;
            });

        return users as UserData[];
    }

    /**
     * Updates a user's data in the database. Only the email and emailValidated fields can be updated.
     *
     * @param object - The updated user data.
     * @returns A promise that resolves to the updated user data.
     */
    async update(user: UserData): Promise<UserData> {
        const sql =
            "UPDATE users SET `email` = ?, `emailValidated` = ? WHERE `id` = ?";

        await pool
            .query(sql, [user.email, user.emailValidated || 0, user.id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Update failed, no rows affected");
                }
            });

        user.last_updated = new Date();

        return user;
    }

    /**
     * Deletes a user from the database.
     * @param id - The ID of the user to delete.
     * @returns A Promise that resolves to a boolean indicating whether the deletion was successful.
     */
    async delete(id: number): Promise<boolean> {
        const sql = "DELETE FROM users WHERE id = ?";

        const success: boolean = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Delete failed, no rows affected");
                }
                return true;
            });

        return success;
    }

    /**
     * Signs up a user with the provided email and password.
     *
     * @param email - The email of the user.
     * @param password - The password of the user.
     * @returns A Promise that resolves to the created user data, or undefined if the user already exists or an error occurs.
     */
    async signup(email: string, password: string): Promise<UserData> {
        // Check if user already exists
        const sql_ex = "SELECT id FROM users WHERE email = ?";
        const exists = await pool
            .query(sql_ex, [email])
            .then(check_mariadb_warning)
            .then((res) => {
                return res.length > 0;
            });

        if (exists) {
            throw new Error("User already exists, cant signup again");
        }

        // hash password to store in db
        const hashedPassword = await bcrypt.hash(password, 15);
        const userData = this.insert({
            email,
            emailValidated: false,
            hashedPassword,
        });

        return userData;
    }

    /**
     * Verifies the email of a user. I.e. updates the user's emailValidated field to true.
     * @param email - The email of the user to verify.
     * @returns A Promise that resolves to a boolean indicating whether the email was successfully verified.
     */
    async verify(email: string): Promise<boolean> {
        // Update user
        const sql = "UPDATE users SET `emailValidated` = ? WHERE `email` = ?";

        const verified = await pool
            .query(sql, [true, email])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows == 0) {
                    throw new Error("Update failed, no rows affected");
                }
                return true;
            });

        return verified;
    }

    /**
     * Authenticates a user by checking their email and password.
     * @param email - The email of the user.
     * @param password - The password of the user.
     * @returns A Promise that resolves to a boolean indicating whether the login was successful.
     */
    async login(email: string, password: string): Promise<boolean> {
        const sql = "SELECT hashedPassword FROM users WHERE email = ?";

        const validated = await pool
            .query(sql, [email])
            .then(check_mariadb_warning)
            .then((res) => {
                // User not found
                if (res.length != 1) {
                    return false;
                }
                return bcrypt.compare(password, res[0].hashedPassword);
            });

        return validated;
    }

    async loginById(id: number, password: string): Promise<boolean> {
        const sql = "SELECT hashedPassword FROM users WHERE id = ?";

        const validated = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                // User not found
                if (res.length != 1) {
                    return false;
                }
                return bcrypt.compare(password, res[0].hashedPassword);
            });

        return validated;
    }

    /**
     * Updates the password for a user.
     *
     * @param email - The email of the user.
     * @param password - The new password to be set.
     * @returns A promise that resolves to a boolean indicating whether the password was successfully updated.
     */
    async updatePassword(email: string, password: string): Promise<boolean> {
        // hash password to store in db
        const hashedPassword = await bcrypt.hash(password, 15);

        const sql = "UPDATE users SET `hashedPassword` = ? WHERE `email` = ?";

        const updated = await pool
            .query(sql, [hashedPassword, email])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows != 1) {
                    throw new Error("Update failed, no rows affected");
                }
                return true;
            });

        return updated;
    }

    async getByEmail(email: string): Promise<UserData> {
        const sql =
            "SELECT  id, email, emailValidated, created, last_updated, last_login, last_login_ip FROM users WHERE email = ?";

        const user = await pool
            .query(sql, [email])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new NotFoundError("user", "email:" + email);
                }
                return res[0];
            });

        return user as UserData;
    }

    async isAdmin(user_id: number) {
        const groups = await groupService.getGroupIdsByUserId(user_id);
        return groups.includes(1); //Admin group id = 1
    }

    async search(search: string): Promise<UserData[]> {
        const sql = `
        SELECT  id, email, emailValidated, created, last_updated, last_login, last_login_ip FROM users
            WHERE email LIKE ? OR
            id LIKE ? OR
            emailValidated LIKE ? OR
            created LIKE ?
        `;

        const users = await pool
            .query(sql, [
                `%${search}%`,
                `%${search}%`,
                `%${search}%`,
                `%${search}%`,
            ])
            .then(check_mariadb_warning)
            .then((res) => {
                return res;
            });

        return users;
    }

    async getPasswordHash(email: string): Promise<string> {
        const sql = "SELECT hashedPassword FROM users WHERE email = ?";

        const hash = await pool
            .query(sql, [email])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new Error(
                        "Number of results does not match! length: " +
                            res.length +
                            " for email: " +
                            email,
                    );
                }
                return res[0].hashedPassword;
            });

        return hash;
    }

    async updateLoginInfo(user_id: number, ip: string): Promise<boolean> {
        const sql =
            "UPDATE users SET `last_login_ip` = ?, `last_login` = CURRENT_TIMESTAMP() WHERE `id` = ?";

        const updated = await pool
            .query(sql, [ip, user_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows != 1) {
                    throw new Error("Update failed, no rows affected");
                }
                return true;
            });

        return updated;
    }
}

/** We need to globally register each service here
 * to prevent memory leaks in development.
 */
export const userService = registerService("db_user", () => new UserService());
